import logging
from typing import Type

from .base import BankBase
from .citibank import Citibank
from .dbs import Dbs
from .detector import BankDetector
from .example_bank import ExampleBank
from .hsbc import Hsbc
from .maybank import Maybank
from .ocbc import Ocbc
from .standard_chartered import StandardChartered
from .uob import Uob
from .zkb import ZurcherKantonalBank

banks: list[Type["BankBase"]] = [
    Citibank,
    Dbs,
    ExampleBank,
    Hsbc,
    Maybank,
    Ocbc,
    StandardChartered,
    Uob,
    ZurcherKantonalBank,
]

logger = logging.getLogger(__name__)

__all__ = ["BankDetector", "BankBase", *[bank.__name__ for bank in banks]]



"""This file stores statement-related enums and constants"""

from enum import auto

from strenum import StrEnum

from monopoly.enums import AutoEnum, RegexEnum

from .date import ISO8601


class EntryType(AutoEnum):
    CREDIT = auto()
    DEBIT = auto()


class BankNames(AutoEnum):
    CITIBANK = auto()
    DBS = auto()
    HSBC = auto()
    MAYBANK = auto()
    OCBC = auto()
    STANDARD_CHARTERED = auto()
    UOB = auto()
    ZKB = auto()


class InternalBankNames(AutoEnum):
    EXAMPLE = auto()
    GENERIC = auto()


class Columns(AutoEnum):
    AMOUNT = auto()
    DATE = auto()
    DESCRIPTION = auto()
    SUFFIX = auto()
    TRANSACTION_DATE = auto()


class SharedPatterns(StrEnum):
    """
    AMOUNT matches the following patterns:
    1,123.12 | 123.12 | (123.12) | ( 123.12) | 123.12 CR

    AMOUNT_EXTENDED is generally used for credit statements and to
    find statement balances, while AMOUNT_EXTENDED_WITHOUT_EOL is used
    for debit statements, since debit statements have the amount + balances
    on a single line, and we only want the amount.
    """

    COMMA_FORMAT = r"\d{1,3}(,\d{3})*\.\d*"
    ENCLOSED_COMMA_FORMAT = rf"\({COMMA_FORMAT}\s{{0,1}}\))"
    DEBIT_CREDIT_SUFFIX = r"(?P<suffix>CR\b|DR\b|\+|\-)?\s*"

    AMOUNT = rf"(?P<amount>{COMMA_FORMAT}|{ENCLOSED_COMMA_FORMAT}\s*"
    AMOUNT_EXTENDED_WITHOUT_EOL = AMOUNT + DEBIT_CREDIT_SUFFIX
    AMOUNT_EXTENDED = AMOUNT_EXTENDED_WITHOUT_EOL + r"$"

    BALANCE = rf"(?P<balance>{COMMA_FORMAT})?$"
    DESCRIPTION = r"(?P<description>.*?)\s+"
    TRANSACTION_DATE_ABBREVIATED_ALL_CAPS = r"(?P<transaction_date>\d{2}\s[A-Z]{3})\s+"
    TRANSACTION_DATE_ABBREVIATED_PROPER_CASE = (
        r"(?P<transaction_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
    )
    POSTING_DATE_ABBREVIATED_PROPER = r"(?P<posting_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"


class StatementBalancePatterns(RegexEnum):
    DBS = (
        r"(?P<description>PREVIOUS BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    CITIBANK = (
        r"(?P<description>BALANCE PREVIOUS STATEMENT?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    HSBC = (
        r"(?P<description>Previous Statement Balance?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    MAYBANK = (
        r"(?P<description>YOUR PREVIOUS STATEMENT BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    OCBC = (
        r"(?P<description>LAST MONTH'S BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    STANDARD_CHARTERED = (
        r"(?P<description>BALANCE FROM PREVIOUS STATEMENT?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    UOB = (
        r"(?P<description>PREVIOUS BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )


class CreditTransactionPatterns(RegexEnum):
    DBS = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    CITIBANK = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    HSBC = (
        rf"(?P<posting_date>{ISO8601.DD_MMM_RELAXED})\s+"
        rf"(?P<transaction_date>{ISO8601.DD_MMM_RELAXED})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    MAYBANK = (
        rf"(?P<posting_date>{ISO8601.DD_MM})\s+"
        rf"(?P<transaction_date>{ISO8601.DD_MM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    OCBC = (
        r"(?P<transaction_date>\d+/\d+)\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    STANDARD_CHARTERED = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        rf"(?P<posting_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?:(?P<transaction_ref>Transaction\sRef\s\d+)?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED
    )
    UOB = (
        rf"(?P<posting_date>{ISO8601.DD_MMM})\s+"
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )


class DebitTransactionPatterns(RegexEnum):
    DBS = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    DBS_POSB_CONSOLIDATED = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YYYY})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    MAYBANK = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YY})\s+"
        + SharedPatterns.DESCRIPTION
        # remove *\s
        + SharedPatterns.AMOUNT[:-3]
        + r"(?P<suffix>\-|\+)\s+"
        + SharedPatterns.BALANCE
    )
    OCBC = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        rf"(?P<posting_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        + SharedPatterns.BALANCE
    )
    UOB = (
        rf"(?P<transaction_date>{ISO8601.DD_MMM})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        + SharedPatterns.BALANCE
    )
    ZKB = (
        rf"(?P<transaction_date>{ISO8601.DD_MM_YYYY})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?P<amount>\d{1,3}(\'\d{3})*(\.\d+)?)\s+"
        + rf"(?P<value_date>{ISO8601.DD_MM_YYYY})\s+"
        + r"(?P<balance>\d{1,3}(\'\d{3})*(\.\d+)?)$"
    )



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Ocbc(BankBase):
    name = BankNames.OCBC

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=ISO8601.DD_MM_YYYY,
        header_pattern=regex(r"(TRANSACTION DATE.*DESCRIPTION.*AMOUNT)"),
        prev_balance_pattern=StatementBalancePatterns.OCBC,
        transaction_pattern=CreditTransactionPatterns.OCBC,
    )

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=regex(rf"\s{ISO8601.DD_MMM_YYYY}$"),
        header_pattern=regex(r"(Withdrawal.*Deposit.*Balance)"),
        transaction_pattern=DebitTransactionPatterns.OCBC,
        multiline_transactions=True,
        transaction_bound=170,
    )

    identifiers = [
        [
            MetadataIdentifier(
                creator="pdfgen", producer="Streamline PDFGen for OCBC Group"
            ),
            TextIdentifier("OCBC"),
        ],
    ]

    statement_configs = [credit, debit]



from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import EntryType, InternalBankNames, SharedPatterns
from monopoly.identifiers import TextIdentifier

from .base import BankBase


class ExampleBank(BankBase):
    """Dummy class to help with reading the example PDF statement"""

    name = InternalBankNames.EXAMPLE

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=regex(r"(\d{2}\-\d{2}\-\d{4})"),
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        prev_balance_pattern=regex(
            r"(?P<description>LAST MONTH'S BALANCE?)\s+"
            + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        ),
        transaction_pattern=regex(
            r"(?P<transaction_date>\d+/\d+)\s*"
            + SharedPatterns.DESCRIPTION
            + SharedPatterns.AMOUNT_EXTENDED
        ),
    )

    identifiers = [
        [
            TextIdentifier(text="HCBS12345(12345)"),
        ]
    ]

    statement_configs = [credit]



from dataclasses import fields
from functools import lru_cache

from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class Identifier:
    """Parent class for identifiers, used for type hinting"""


@dataclass(frozen=True)
class MetadataIdentifier(Identifier):
    """Stores the metadata attributes of a PDF"""

    format: str = ""
    title: str = ""
    author: str = ""
    subject: str = ""
    creator: str = ""
    producer: str = ""

    @lru_cache
    def matches(self, other: "MetadataIdentifier") -> bool:
        """Check for partial matches on all string fields."""
        for field in fields(self):
            self_value = getattr(self, field.name)
            other_value = getattr(other, field.name)

            # Perform partial matching if both fields are non-empty strings
            if isinstance(self_value, str) and isinstance(other_value, str):
                if other_value and other_value not in self_value:
                    return False

        return True


@dataclass(frozen=True)
class TextIdentifier(Identifier):
    """Stores a specific string that exists in the content of a PDF"""

    text: str = ""

    @lru_cache
    def matches(self, raw_text: str) -> bool:
        """Check for partial matches on all string fields."""
        return self.text in raw_text



import logging
from typing import Any

from monopoly.config import PdfConfig, StatementConfig

logger = logging.getLogger(__name__)


class BankBase:
    """
    Abstract class to handle initialization of common variables
    that are shared between bank processor classes.

    Ensures consistency between bank classes.
    """

    name: str
    statement_configs: list[StatementConfig]
    pdf_config: PdfConfig = PdfConfig()
    identifiers: list[list[Any]]

    def __init_subclass__(cls, **kwargs) -> None:
        if not hasattr(cls, "statement_configs"):
            raise NotImplementedError(
                f"{cls.__class__.__name__} "
                "must implement `statement_configs` class variable"
            )
        if not hasattr(cls, "identifiers"):
            raise NotImplementedError(
                f"{cls.__class__.__name__} "
                "must implement `identifiers` class variable"
            )
        if not hasattr(cls, "name"):
            raise NotImplementedError(
                f"{cls.__class__.__name__} " "must implement `name` class variable"
            )

        # validation logic only applies to regular banks
        if cls.identifiers:
            if not any(isinstance(item, list) for item in cls.identifiers):
                raise TypeError("`identifiers` must be a list of lists")

        return super().__init_subclass__(**kwargs)



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Citibank(BankBase):
    name = BankNames.CITIBANK

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=regex(r"Statement\sDate\s+(.*)"),
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        prev_balance_pattern=StatementBalancePatterns.CITIBANK,
        transaction_pattern=CreditTransactionPatterns.CITIBANK,
    )

    identifiers = [
        [
            MetadataIdentifier(
                creator="Ricoh Americas Corporation, AFP2PDF",
                producer="Ricoh Americas Corporation, AFP2PDF",
            )
        ]
    ]

    statement_configs = [credit]



from .generic import DateMatch, DatePatternAnalyzer
from .handler import GenericBank, GenericStatementHandler

__all__ = ["DatePatternAnalyzer", "DateMatch", "GenericStatementHandler", "GenericBank"]



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Maybank(BankBase):
    name = BankNames.MAYBANK

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=regex(rf"(?:結單日期)[:\s]+{ISO8601.DD_MM_YY}"),
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT.*BALANCE)"),
        transaction_pattern=DebitTransactionPatterns.MAYBANK,
        multiline_transactions=True,
    )

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=ISO8601.DD_MMM_YY,
        header_pattern=regex(r"(Date.*Description.*Amount)"),
        transaction_pattern=CreditTransactionPatterns.MAYBANK,
        prev_balance_pattern=StatementBalancePatterns.MAYBANK,
        multiline_transactions=True,
    )

    identifiers = [
        [
            MetadataIdentifier(
                author="Maybank2U.com",
                creator="Maybank2u.com",
                producer="iText",
            ),
        ],
        [
            MetadataIdentifier(
                title="Credit Card Statement",
                author="Maybank2U.com",
                producer="iText",
            ),
        ],
    ]

    statement_configs = [credit, debit]



import logging
import re
from functools import lru_cache

from monopoly.constants import EntryType
from monopoly.statements.transaction import TransactionMatch

from .base import BaseStatement, SafetyCheckError

logger = logging.getLogger(__name__)


class DebitStatement(BaseStatement):
    """
    A dataclass representation of a debit statement
    """

    statement_type = EntryType.DEBIT

    def pre_process_match(
        self, transaction_match: TransactionMatch
    ) -> TransactionMatch:
        """
        Pre-processes transactions by adding a debit or credit suffix to the group dict
        """
        if self.config.statement_type == EntryType.DEBIT:
            transaction_match.groupdict.suffix = self.get_debit_suffix(
                transaction_match
            )
        return transaction_match

    def get_debit_suffix(self, transaction_match: TransactionMatch) -> str | None:
        """
        Gets the accounting suffix for debit card statements

        Attempts to identify whether a transaction is a debit
        or credit entry based on the distance from the withdrawal
        or deposit columns.
        """
        page_number = transaction_match.page_number
        withdrawal_pos = self.get_withdrawal_pos(page_number)
        deposit_pos = self.get_deposit_pos(page_number)

        if deposit_pos and withdrawal_pos:
            amount = transaction_match.groupdict.amount
            line: str = transaction_match.match.string
            start_pos = line.find(amount)
            # assume that numbers are right aligned
            end_pos = start_pos + len(amount) - 1
            withdrawal_diff = abs(end_pos - withdrawal_pos)
            deposit_diff = abs(end_pos - deposit_pos)
            if withdrawal_diff > deposit_diff:
                return "CR"
            return "DR"
        return transaction_match.groupdict.suffix

    @lru_cache
    def get_withdrawal_pos(self, page_number: int) -> int | None:
        common_names = ["withdraw", "debit"]
        for name in common_names:
            if pos := self.get_column_pos(name, page_number=page_number):
                return pos
        logger.debug(
            "%s column not found in header on page %s", common_names, page_number
        )
        return False

    @lru_cache
    def get_deposit_pos(self, page_number: int) -> int | None:
        common_names = ["deposit", "credit"]
        for name in common_names:
            if pos := self.get_column_pos(name, page_number=page_number):
                return pos
        logger.debug(
            "%s column not found in header on page %s", common_names, page_number
        )
        return False

    @lru_cache
    def get_column_pos(self, column_type: str, page_number: int) -> int | None:
        pattern = re.compile(rf"{column_type}[\w()$]*", re.IGNORECASE)
        if match := pattern.search(self.header):
            return self.get_header_pos(match.group(), page_number)
        return None

    @lru_cache
    def get_header_pos(self, column_name: str, page_number: int) -> int:
        """
        Returns position of the 'WITHDRAWAL' or 'DEPOSIT' header on a bank statement
        for a particular page.

        An assumption is made here that numbers are right aligned, meaning
        that if an amount matches with the end of the withdrawal string position,
        the item is in fact a withdrawal

        e.g.
        ```
        DATE         DESCRIPTION          WITHDRAWAL         DEPOSIT
        15 OCT       bill payment             322.07
        16 OCT       item                                     123.12
        ```
        """
        header_pattern = self.config.header_pattern
        lines = self.pages[page_number].lines
        for line in lines:
            if match := header_pattern.search(line):
                header = match.string.lower()
                header_start_pos = header.find(column_name.lower())
                if header_start_pos == -1:
                    continue
                return header_start_pos + len(column_name)

        raise ValueError(
            f"Debit header {column_name} cannot be found on page {page_number}"
        )

    @lru_cache
    def perform_safety_check(self) -> bool:
        """
        Checks that debit and credit transaction sums
        exist as a number within the statement
        """
        transactions = self.transactions

        numbers = self.get_all_numbers_from_document()

        # add a zero, for cases where the debit statement
        # either is completely debit or credit transactions
        numbers.update([0])

        debit_amounts = [
            transaction.amount for transaction in transactions if transaction.amount > 0
        ]
        credit_amounts = [
            transaction.amount for transaction in transactions if transaction.amount < 0
        ]

        debit_sum = round(abs(sum(debit_amounts)), 2)
        credit_sum = round(abs(sum(credit_amounts)), 2)

        result = all([debit_sum in numbers, credit_sum in numbers])
        if not result:
            raise SafetyCheckError(self.failed_safety_message)

        return result



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import ISO8601, BankNames, DebitTransactionPatterns, EntryType
from monopoly.identifiers import MetadataIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class ZurcherKantonalBank(BankBase):
    name = BankNames.ZKB

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=regex(rf"Balance as of: ({ISO8601.DD_MM_YYYY})"),
        header_pattern=regex(r"(Date.*Booking text.*Debit CHF.*Credit CHF)"),
        transaction_pattern=DebitTransactionPatterns.ZKB,
        multiline_transactions=True,
        safety_check=False,
    )

    identifiers = [
        [
            MetadataIdentifier(
                format="PDF 1.7",
                title="SLK_Vermoegensinfo_Group",
                creator="Designer",
                producer="PDFlib+PDI",
            )
        ]
    ]
    statement_configs = [debit]



"""This file stores date-related regex patterns and constants"""

import logging

from strenum import StrEnum

from monopoly.enums import RegexEnum

logger = logging.getLogger(__name__)


# flake8: noqa
# pylint: disable=line-too-long
class DateFormats(StrEnum):
    """Holds a case-insensitive list of common ISO 8601 date formats"""

    D = r"(1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)"
    DD = r"(01|02|03|04|05|06|07|08|09|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)"
    M = r"(1|2|3|4|5|6|7|8|9|10|11|12)"
    MM = r"(01|02|03|04|05|06|07|08|09|10|11|12)"
    MMM = r"(?i:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"
    MMMM = r"(?i:January|February|March|April|May|June|July|August|September|October|November|December)"
    YY = r"([2-5][0-9]\b)"
    YYYY = r"(20\d{2}\b)"


class ISO8601(RegexEnum):
    DD_MM = rf"\b({DateFormats.DD}[\/\-\s]{DateFormats.MM})"
    DD_MM_YY = rf"\b({DateFormats.DD}[\/\-\s]{DateFormats.MM}[\/\-\s]{DateFormats.YY})"
    DD_MM_YYYY = (
        rf"\b({DateFormats.DD}[\/\-\s.]{DateFormats.MM}[\/\-\s.]{DateFormats.YYYY})"
    )
    DD_MMM = rf"\b({DateFormats.DD}[-\s]{DateFormats.MMM})"
    DD_MMM_RELAXED = DD_MMM.replace(r"[-\s]", r"(?:[-\s]|)")
    DD_MMM_YY = rf"\b({DateFormats.DD}[-\s]{DateFormats.MMM}[-\s]{DateFormats.YY})"
    DD_MMM_YYYY = (
        rf"\b({DateFormats.DD}[-\s]{DateFormats.MMM}[,\s]{{1,2}}{DateFormats.YYYY})"
    )
    MMMM_DD_YYYY = (
        rf"\b({DateFormats.MMMM}\s{DateFormats.DD}[,\s]{{1,2}}{DateFormats.YYYY})"
    )
    MMM_DD = rf"\b({DateFormats.MMM}[-\s]{DateFormats.DD})"
    MMM_DD_YYYY = (
        rf"\b({DateFormats.MMM}[-\s]{DateFormats.DD}[,\s]{{1,2}}{DateFormats.YYYY})"
    )



import csv
import logging
from pathlib import Path
from typing import Optional, Type

from dateparser import parse
from pydantic import SecretStr

from monopoly.banks import BankBase
from monopoly.config import DateOrder
from monopoly.generic import GenericBank, GenericStatementHandler
from monopoly.handler import StatementHandler
from monopoly.pdf import PdfPage, PdfParser
from monopoly.statements import BaseStatement, Transaction
from monopoly.write import generate_name

logger = logging.getLogger(__name__)


class Pipeline:
    """Handles extract, transform and load (ETL) logic for bank statements"""

    def __init__(
        self,
        parser: PdfParser,
        passwords: Optional[list[SecretStr]] = None,
    ):
        self.passwords = passwords
        pages = parser.get_pages()
        self.handler = self.create_handler(parser.bank, pages)

    @staticmethod
    def create_handler(bank: Type[BankBase], pages: list[PdfPage]) -> StatementHandler:
        if issubclass(bank, GenericBank):
            logger.debug("Using generic statement handler")
            return GenericStatementHandler(bank, pages)
        logger.debug("Using statement handler with bank: %s", bank.__name__)
        return StatementHandler(bank, pages)

    def extract(self, safety_check=True) -> BaseStatement:
        """Extracts transactions from the statement, and performs
        a safety check to make sure that total transactions add up"""
        statement = self.handler.get_statement()
        transactions = statement.get_transactions()

        if not transactions:
            raise ValueError("No transactions found - statement extraction failed")

        logger.debug("%s transactions found", len(transactions))

        if not statement.statement_date:
            raise ValueError("No statement date found")

        if safety_check and statement.config.safety_check:
            statement.perform_safety_check()

        return statement

    @staticmethod
    def transform(statement: BaseStatement) -> list[Transaction]:
        logger.debug("Running transformation functions on DataFrame")
        transactions = statement.transactions
        statement_date = statement.statement_date
        transaction_date_order = statement.config.transaction_date_order

        def convert_date(transaction: Transaction, transaction_date_order: DateOrder):
            """
            Converts each date to a ISO 8601 (YYYY-MM-DD) format.

            Implements cross-year logic by attributing transactions from
            October, November, and December to the previous year if
            the statement month is January.
            e.g. if the statement month is Jan/Feb 2024, transactions from
            Oct/Nov/Dec should be attributed to the previous year.
            """
            if str(statement_date.year) not in transaction.date:
                transaction.date += f" {statement_date.year}"
            parsed_date = parse(
                transaction.date,
                settings=transaction_date_order.settings,
            )
            if parsed_date:
                if statement_date.month in (1, 2) and parsed_date.month > 2:
                    parsed_date = parsed_date.replace(year=parsed_date.year - 1)

                return parsed_date.isoformat()[:10]
            raise RuntimeError("Could not convert date")

        logger.debug("Transforming dates to ISO 8601")

        for transaction in transactions:
            transaction.date = convert_date(transaction, transaction_date_order)
        return transactions

    @staticmethod
    def load(
        transactions: list[Transaction],
        statement: BaseStatement,
        output_directory: Path,
    ):
        if isinstance(output_directory, str):
            output_directory = Path(output_directory)

        filename = generate_name(
            statement=statement,
            format_type="file",
            bank_name=statement.bank_name,
            statement_type=statement.statement_type,
            statement_date=statement.statement_date,
        )

        output_path = output_directory / filename
        logger.debug("Writing CSV to file path: %s", output_path)

        with open(output_path, mode="w", encoding="utf8") as file:
            writer = csv.writer(file)

            # header
            writer.writerow(statement.columns)

            for transaction in transactions:
                writer.writerow(
                    (
                        [
                            transaction.date,
                            transaction.description,
                            transaction.amount,
                        ]
                    )
                )

        return output_path



import logging
from re import compile as regex

from monopoly.config import PdfConfig, StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Hsbc(BankBase):
    name = BankNames.HSBC

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=regex(rf"Statement From .* to {ISO8601.DD_MMM_YYYY}"),
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        prev_balance_pattern=StatementBalancePatterns.HSBC,
        transaction_pattern=CreditTransactionPatterns.HSBC,
        multiline_transactions=True,
    )

    email_statement_identifier = [
        MetadataIdentifier(
            title="PRJ_BEAGLE_ST_CNS_SGH_APP_Orchid",
            author="Registered to: HSBCGLOB",
            creator="OpenText Exstream",
        ),
        TextIdentifier("HSBC"),
    ]

    web_and_mobile_statement_identifier = [
        MetadataIdentifier(
            format="PDF 1.7", producer="OpenText Output Transformation Engine"
        )
    ]

    pdf_config = PdfConfig(
        page_bbox=(0, 0, 379, 840),
        ocr_identifiers=web_and_mobile_statement_identifier,
    )

    identifiers = [email_statement_identifier, web_and_mobile_statement_identifier]

    statement_configs = [credit]



import logging
from typing import TYPE_CHECKING, Type

from monopoly.identifiers import MetadataIdentifier, TextIdentifier
from monopoly.pdf import PdfDocument

if TYPE_CHECKING:
    from .base import BankBase

logger = logging.getLogger(__name__)


class BankDetector:
    def __init__(self, document: PdfDocument):
        self.document = document
        self.metadata_identifier = document.metadata_identifier

    def detect_bank(self, banks: list[Type["BankBase"]]) -> Type["BankBase"] | None:
        """
        Reads the encryption metadata or actual metadata (if the PDF is not encrypted),
        and checks for a bank based on unique identifiers.
        """
        if not banks:
            banks = []

        logger.debug("Found PDF properties: %s", self.metadata_identifier)

        for bank in banks:
            if self.is_bank_identified(bank):
                return bank
        return None

    def is_bank_identified(
        self,
        bank: Type["BankBase"],
    ) -> bool:
        """
        Checks if a bank is identified based on a list of metadata items.
        """
        for grouped_identifiers in bank.identifiers:  # type: ignore
            if self.identifiers_match(grouped_identifiers):
                logger.debug("Identified statement bank: %s", bank.__name__)
                return True

        return False

    def text_identifiers_match(self, text_identifiers: list[TextIdentifier]) -> bool:
        if not text_identifiers:
            return True

        for identifier in text_identifiers:
            if not identifier.matches(self.document.raw_text):
                return False

        logger.debug("Text identifier found in PDF")
        return True

    def metadata_identifiers_match(
        self, bank_metadata_identifiers: list[MetadataIdentifier]
    ) -> bool:
        for identifier in bank_metadata_identifiers:
            if self.metadata_identifier.matches(identifier):
                return True
        return False

    def identifiers_match(self, identifiers: list) -> bool:
        text_identifiers = self.get_identifiers_of_type(identifiers, TextIdentifier)
        metadata_identifiers = self.get_identifiers_of_type(
            identifiers, MetadataIdentifier
        )

        if metadata_identifiers:
            if not self.metadata_identifiers_match(metadata_identifiers):
                return False
            if not self.text_identifiers_match(text_identifiers):
                return False

        return self.text_identifiers_match(text_identifiers)

    def get_identifiers_of_type(self, identifiers: list, identifier_type: Type) -> list:
        return [i for i in identifiers if isinstance(i, identifier_type)]



import logging
from re import compile as regex

from monopoly.config import StatementConfig
from monopoly.constants import (
    ISO8601,
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    EntryType,
    StatementBalancePatterns,
)
from monopoly.identifiers import MetadataIdentifier, TextIdentifier

from ..base import BankBase

logger = logging.getLogger(__name__)


class Dbs(BankBase):
    name = BankNames.DBS

    credit = StatementConfig(
        statement_type=EntryType.CREDIT,
        statement_date_pattern=ISO8601.DD_MMM_YYYY,
        multiline_transactions=False,
        header_pattern=regex(r"(DATE.*DESCRIPTION.*AMOUNT)"),
        transaction_pattern=CreditTransactionPatterns.DBS,
        prev_balance_pattern=StatementBalancePatterns.DBS,
    )

    debit = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=ISO8601.DD_MMM_YYYY,
        multiline_transactions=True,
        header_pattern=regex(r"(WITHDRAWAL.*DEPOSIT.*BALANCE)"),
        transaction_pattern=DebitTransactionPatterns.DBS,
        transaction_bound=170,
    )

    consolidated = StatementConfig(
        statement_type=EntryType.DEBIT,
        statement_date_pattern=regex(rf"Details as at {ISO8601.DD_MMM_YYYY}"),
        multiline_transactions=True,
        header_pattern=regex(r"(\s+Date\s+Description\s+Withdrawal \(-\).*)"),
        transaction_pattern=DebitTransactionPatterns.DBS_POSB_CONSOLIDATED,
        transaction_bound=100,
    )

    identifiers = [
        [
            TextIdentifier("DBS"),
            MetadataIdentifier(creator="Quadient CXM AG"),
        ],
    ]

    statement_configs = [credit, consolidated, debit]
